
from __future__ import print_function

from sklearn.base import BaseEstimator, ClassifierMixin

import numpy

class KernelClassifier(BaseEstimator, ClassifierMixin):

    def find_classes(self, Y):
        classes = list(set(Y))
        if len(classes) != 2:
            raise Exception("this does not seem to be a 2-class problem")
        self.positive_class = classes[0]
        self.negative_class = classes[1]

    def compute_score(self, x):
        return sum( a_i * self.K(x_i, x) 
                    for a_i, x_i in zip(self.alphas, self.xs) )
        
    def single_predict(self, x):
        score = self.compute_score(x)
        return self.positive_class if score > 0 else self.negative_class

    def predict(self, X):
        return [ self.single_predict(x) for x in X ]

kernel = 'polynomial'

def my_kernel(x1, x2):
    if kernel == 'linear':
        return numpy.dot(x1, x2)
    if kernel == 'polynomial':
        offset = 1
        degree = 2
        return (numpy.dot(x1, x2) + offset) ** degree
    if kernel == 'rbf':
        gamma = 1
        return numpy.exp( -gamma * numpy.dot(x1-x2, x1-x2)**2 )

class KernelPerceptron(KernelClassifier):

    def __init__(self, n_iter=5):
        self.n_iter = n_iter
        self.K = my_kernel

    def fit(self, X, Y):
        Y = list(Y)
        self.find_classes(Y)        
        self.xs = list(X)
        self.alphas = [ 0 for _ in range(len(Y)) ]
        for i in range(self.n_iter):            
            for i, (x, y) in enumerate(zip(X, Y)):                
                score = self.compute_score(x)
                if score < 0 and y == self.positive_class:
                    self.alphas[i] += 1
                elif score >= 0 and y == self.negative_class:
                    self.alphas[i] -= 1


